local Entity = import("/lua/sim/entity.lua").Entity
local utilities = import('/lua/utilities.lua')

-- Log the step by step process at aid in troubleshooting
local myDebug = false

-- Test Mode allows the player to see the 'Radar' cone by making it 'Vision' instead
local testMode = true

local oldIntelComponent = IntelComponent
IntelComponent = Class(oldIntelComponent) {

    ---@param self IntelComponent | Unit
    ---@param builder Unit unused
    ---@param layer Layer unused
    OnStopBeingBuilt = function(self, builder, layer)
        local intelBlueprint = self.Blueprint.Intel
        if intelBlueprint and intelBlueprint.State then
            self.IntelStatus = table.deepcopy(intelBlueprint.State)
            self:EnableUnitIntel('NotInitialized')
            self.Brain:AddEnergyDependingEntity(self)
        end    
    end,

    ---@param self IntelComponent | Unit
    OnEnergyDepleted = function(self)
        local status = self.IntelStatus
        if status then
            self:DisableUnitIntel('Energy')
        end
    end,

    ---@param self IntelComponent | Unit
    OnEnergyViable = function(self)
        local status = self.IntelStatus
        if status then
            self:EnableUnitIntel('Energy')
        end
    end,
    
-----------------------------------------------------------------------------------------------------------------------
--	Start of "Radar Cone" scripts, Creates a 45 degree cone of intel, originating out from the parent units chosen bone
-----------------------------------------------------------------------------------------------------------------------    

InitializeRadarCone = function(self)
	if myDebug then WARN('InitializeRadarCone') end
	
    --	Int the entities table
	self.RadarConeEntitiesTable = {}
		
	-- Set up local variables from spec
	local bp = self:GetBlueprint()
	local bone
	local radius = bp.Intel.RadarRadius
	local rotate = false
	
	-- Find the highest bone for Structures and Naval units
	if table.find(bp.Categories ,'STRUCTURE') or table.find(bp.Categories ,'NAVAL') then
		rotate = self:RtnRotator()
		if not bone then bone = self:RtnHighestBone() end
	else
		-- Give non Structure / non Naval units a boost in range as their cone is narrow
		radius = radius * 2
	end 		
		
	if myDebug then WARN('	bone: ', bone,'	radius: ', radius,'	rotate: ', rotate) end
	if radius > 0 then
		local counter = 1
		local loopFlag = true
		while not self:BeenDestroyed() and counter <= radius and loopFlag do
			-- Create a radar arc thats approx 45 degrees and is roughly the radius needed
			local angle
			if bone and rotate then
				-- Units with rotating cone, have a wider cone
				angle = math.ceil((counter / 3.14) * 2)
			else
				-- units with non-rotating cone have a narrow one
				angle = math.ceil((counter / 3.14) * 1)
			end
			if counter + angle < radius then
				if myDebug then WARN('	Counter + angle < radius') end
				-- Attach the entity to parent
				local radEnt = import('/lua/sim/Entity.lua').Entity({Owner = self, Name = counter,})
				if myDebug then WARN('	entity created') end
				radEnt:AttachBoneTo( -1, self, bone or 0 )
				if myDebug then WARN('	entity attached') end
				table.insert(self.RadarConeEntitiesTable, radEnt)
				if myDebug then WARN('	entity inserted into tbl: ', table.getn(self.RadarConeEntitiesTable)) end
				self.Trash:Add(radEnt)
				if myDebug then WARN('	entity added to trash') end
				--	Offset the entities so they create a cone from the parent
				radEnt:SetParentOffset(Vector(0,0, counter))
				if myDebug then WARN('	entity offset') end
				-- Set who can view the entities intel
				radEnt:SetVizToFocusPlayer('Always')
				radEnt:SetVizToAllies('Always')
				radEnt:SetVizToNeutrals('Never')
				radEnt:SetVizToEnemies('Never')
				if myDebug then WARN('	focus set') end
				-- Initalize intel
				if testMode then
					radEnt:InitIntel(self:GetArmy(), 'Vision', angle)
				else
					radEnt:InitIntel(self:GetArmy(), 'Radar', angle) 
				end
				if myDebug then WARN('	intel initialized') end
	        else
	        	loopFlag = false
	        	if myDebug then WARN('	flipping flag to end loop early') end
	        end
	        if loopFlag then
				counter = counter + 1
				if myDebug then WARN('	counter increased') end
			end
	    end
		if myDebug then WARN('	end loop') end
		--	Initialize and enable intel types
		self:EnableRadarCone()
	    if bone and rotate and not self.RadarConeRotator then
			if myDebug then WARN('	rotating') end
	    	-- Rotate the radar cone if a bone is provided
	    	local rotation
	    	if table.find(bp.Categories ,'TECH1') then
	    		rotation = 36
	    	elseif table.find(bp.Categories ,'TECH2') then
	    		rotation = 24
	    	else
	    		rotation = 12
	    	end
	    	if Random(1,2) == 1 then
	    		rotation = rotation
	    	else
	    		rotation = rotation * -1
	    	end
	    	self.RadarConeRotator = CreateRotator(self, bone, 'y', nil, 0, 999, rotation)
	    	self.Trash:Add(self.RadarConeRotator)
	    end

	end
end,

RtnRotator = function(self)
	-- I wish there was a better way of doing this
	if self.LowerRotator then return false		
	elseif self.Dish1Rotator then return false		
	elseif self.Dish2Rotator then return false		
	elseif self.Dish3Rotator then return false		
	elseif self.MainRotator then return false		
	elseif self.Rotator then return false
	elseif self.Rotator1 then return false
	elseif self.Rotator2 then return false
	elseif self.Rotator3 then return false				
	elseif self.MySpinner then return false	
	elseif self.Spinner01 then return false	
	elseif self.Spinner02 then return false
	elseif self.UpperRotator then return false	
	else
		return true	
	end
end,

RtnHighestBone = function(self)
	local boneTbl = {}
	for i = 1, self:GetBoneCount() do
		if self:IsValidBone(i) then table.insert(boneTbl, self:GetBoneName(i) ) end
	end
	local boneDist = {}
	local sX, sY, sZ = unpack(self:GetPosition())
	local surf = GetSurfaceHeight(sX, sZ)
	for c = 1, table.getn(boneTbl) do
		local boneName = self:GetBoneName(c)
		local bS, bY, bZ = unpack(self:GetPosition(boneName))
		table.insert(boneDist, {dist = (sY - bY), bone = c})
	end
	table.sort(boneDist, sort_by('dist'))
	bone = self:GetBoneName(boneDist[1].bone)
	return bone
end,

EnableRadarCone = function(self)
	if myDebug then WARN('EnableRadarCone') end	
	if not self.RadarConeActive and table.getn(self.RadarConeEntitiesTable) > 0  then
   		for k, v in self.RadarConeEntitiesTable do
   			if testMode then
				v:EnableIntel('Vision')
			else
				v:EnableIntel('Radar')		
			end
		end
		-- Allow the radar to activate on the unit but reduce it to "1". 
		-- Possible that the game may need some intel on the unit it thniks the intel will be active for.
		self:SetIntelRadius('Radar', 1)
		self:EnableIntel('Radar')
		self:RequestRefreshUI()
		self.RadarConeActive = true
	end
end,

DisableRadarCone = function(self)
	if myDebug then WARN('DisableRadarCone') end
	if self.RadarConeActive and table.getn(self.RadarConeEntitiesTable) > 0 then
   		for k, v in self.RadarConeEntitiesTable do
   			if testMode then		
				v:DisableIntel('Vision')
			else
				v:DisableIntel('Radar')
			end
		end
		self:DisableIntel('Radar')
		self:RequestRefreshUI()		
		self.RadarConeActive = false
	end
end,

-----------------------------------------------------------------------------------------------------------------------  
--	End of Radar cone scripts
-----------------------------------------------------------------------------------------------------------------------    
    
    ---@param self IntelComponent | Unit
    ---@param disabler string
    ---@param intel? IntelType
    DisableUnitIntel = function(self, disabler, intel)
        local status = self.IntelStatus
        if status then
            if myDebug then WARN("DisableUnitIntel: " .. tostring(disabler) .. " for " .. tostring(intel)) end

            -- prevent recharging from occuring
            self:OnIntelRechargeFailed()

            -- disable all intel
            local allIntel = status.AllIntel
            local allIntelDisabledByEvent = status.AllIntelDisabledByEvent
            local allIntelMaintenanceFree = status.AllIntelMaintenanceFree
            local allIntelFromEnhancements = status.AllIntelFromEnhancements
            if not intel then
                for i, _ in allIntel do
                    if not (disabler == 'Energy' and allIntelMaintenanceFree and allIntelMaintenanceFree[i]) then
                        allIntelDisabledByEvent[i] = allIntelDisabledByEvent[i] or {}
                        if not allIntelDisabledByEvent[i][disabler] then
                            allIntelDisabledByEvent[i][disabler] = true
                            if i != 'Radar' then
                            	self:DisableIntel(i)
                            elseif self.RadarConeActive then
	                            if myDebug then WARN('	disable all intel, NOT allIntelMaintenanceFree') end  
	                            self:DisableRadarCone()
                            end
                            self:OnIntelDisabled(i)
                        end
                    end
                end

                if allIntelMaintenanceFree then
                    for i, _ in allIntelMaintenanceFree do
                        if not (disabler == 'Energy' and allIntelMaintenanceFree and allIntelMaintenanceFree[i]) then
                            allIntelDisabledByEvent[i] = allIntelDisabledByEvent[i] or {}
                            if not allIntelDisabledByEvent[i][disabler] then
                                allIntelDisabledByEvent[i][disabler] = true
                                if i != 'Radar' then
                                	self:DisableIntel(i)
                                elseif self.RadarConeActive then
	                               	if myDebug then WARN('	disable all intel, IS allIntelMaintenanceFree') end  
	                               	self:DisableRadarCone() 
                                end
                                self:OnIntelDisabled(i)
                            end
                        end
                    end
                end

                -- disable one intel
            elseif allIntel[intel] or (allIntelFromEnhancements and allIntelFromEnhancements[intel]) then
                -- special case that requires additional book keeping
                if disabler == 'Enhancement' then
                    allIntelFromEnhancements[intel] = true
                end

                if not (disabler == 'Energy' and allIntelMaintenanceFree and allIntelMaintenanceFree[intel]) then
                    allIntelDisabledByEvent[intel] = allIntelDisabledByEvent[intel] or {}
                    if not allIntelDisabledByEvent[intel][disabler] then
                        allIntelDisabledByEvent[intel][disabler] = true
                        if intel != 'Radar' then
                        	self:DisableIntel(intel)
                        elseif self.RadarConeActive then
							if myDebug then WARN('	disable one intel') end  
							self:DisableRadarCone() 
                        end
                        self:OnIntelDisabled(intel)
                    end
                end
            end
        end
    end,

    ---@param self IntelComponent | Unit
    ---@param disabler string
    ---@param intel? IntelType
    EnableUnitIntel = function(self, disabler, intel)
        local status = self.IntelStatus
        if status then
            if myDebug then WARN("EnableUnitIntel: " .. tostring(disabler) .. " for " .. tostring(intel)) end

            local allIntel = status.AllIntel
            local allIntelDisabledByEvent = status.AllIntelDisabledByEvent
            local allIntelMaintenanceFree = status.AllIntelMaintenanceFree
            local allIntelFromEnhancements = status.AllIntelFromEnhancements

            -- special case when unit is finished building
            if disabler == 'NotInitialized' then

                -- this bit is weird, but unit logic expects to always have intel immediately enabled when
                -- the unit is done constructing, regardless whether the unit is able to use the intel
                for i, _ in allIntel do
                	if myDebug then WARN('	all intel enabled') end
                    self:OnIntelEnabled(i)
                    if i != 'Radar' then
                    	if myDebug then WARN('	Non-Radar intel enabled') end
                    	self:EnableIntel(i)
                    elseif not self.RadarConeEntitiesTable then
                    	if myDebug then WARN('	Calling InitializeRadarCone') end
                    	self:InitializeRadarCone()
                    else
                    	if myDebug then WARN('	Calling EnableRadarCone') end
                    	self:EnableRadarCone()
                    end
                end

                if allIntelMaintenanceFree then
                	if myDebug then WARN(' all intel enabled, MaintenanceFree') end
                    for i, _ in allIntelMaintenanceFree do
                    	if i != 'Radar' then
                    	    if myDebug then WARN('	Non-Radar intel enabled') end
                        	self:EnableIntel(i)
                    	elseif not self.RadarConeEntitiesTable then
                    		if myDebug then WARN('	Calling InitializeRadarCone') end
                    		self:InitializeRadarCone()
                        else
                        	if myDebug then WARN('	Calling EnableRadarCone') end
                        	self:EnableRadarCone()
                        end
                        self:OnIntelEnabled(i)
                    end
                end

                return
            end

            -- disable all intel
            if not intel then
                for i, _ in allIntel do
                    if not (disabler == 'Energy' and allIntelMaintenanceFree and allIntelMaintenanceFree[i]) then
                        allIntelDisabledByEvent[i] = allIntelDisabledByEvent[i] or {}
                        if allIntelDisabledByEvent[i][disabler] then
                            allIntelDisabledByEvent[i][disabler] = nil
                            if table.empty(allIntelDisabledByEvent[i]) then
                                self:OnIntelRecharge(i)
                            end
                        end
                    end
                end

                if allIntelFromEnhancements then
                    for i, _ in allIntelFromEnhancements do
                        if not (disabler == 'Energy' and allIntelMaintenanceFree and allIntelMaintenanceFree[i]) then
                            allIntelDisabledByEvent[i] = allIntelDisabledByEvent[i] or {}
                            if allIntelDisabledByEvent[i][disabler] then
                                allIntelDisabledByEvent[i][disabler] = nil
                                if table.empty(allIntelDisabledByEvent[i]) then
                                    self:OnIntelRecharge(i)
                                end
                            end
                        end
                    end
                end

                -- disable one intel
            elseif allIntel[intel] or (allIntelFromEnhancements and allIntelFromEnhancements[intel]) then
                -- special case that requires additional book keeping
                if disabler == 'Enhancement' then
                    allIntelFromEnhancements[intel] = true
                end

                if not (disabler == 'Energy' and allIntelMaintenanceFree and allIntelMaintenanceFree[intel]) then
                    allIntelDisabledByEvent[intel] = allIntelDisabledByEvent[intel] or {}
                    if allIntelDisabledByEvent[intel][disabler] then
                        allIntelDisabledByEvent[intel][disabler] = nil
                        if table.empty(allIntelDisabledByEvent[intel]) then
                            self:OnIntelRecharge(intel)
                        end
                    end
                end
            end
        end
    end,

    ---@param self IntelComponent | Unit
    ---@param intel IntelType
    OnIntelRecharge = function(self, intel)
        local status = self.IntelStatus
        if status then
            if myDebug then WARN("OnIntelRecharge: for " .. tostring(intel)) end
            if not status.RechargeThread then
                status.RechargeThread = ForkThread(self.IntelRechargeThread, self)
            end

            status.AllIntelRecharging[intel] = true
        end
    end,

    ---@param self IntelComponent | Unit
    OnIntelRecharged = function(self)
        local status = self.IntelStatus
        if status and status.RechargeThread then
            status.RechargeThread = nil
            for i, _ in status.AllIntelRecharging do
                
            
				-- Allow for the Cone Radar to reactivate after recharge period
            	if i != 'Radar' then
            	    if myDebug then WARN('	Non-Radar intel enabled') end
                	self:EnableIntel(i)
            	elseif not self.RadarConeEntitiesTable then
            		if myDebug then WARN('	Calling InitializeRadarCone') end
            		self:InitializeRadarCone()
                else
                	if myDebug then WARN('	Calling EnableRadarCone') end
                	self:EnableRadarCone()
                end                
                               
                self:OnIntelEnabled(i)
                status.AllIntelRecharging[i] = nil
            end
        end
    end,

    ---@param self IntelComponent | Unit
    OnIntelRechargeFailed = function(self)
        local status = self.IntelStatus
        if status and status.RechargeThread then
            status.RechargeThread:Destroy()
            status.RechargeThread = nil
            self:SetWorkProgress(0)
        end
    end,

    ---@param self IntelComponent | Unit
    IntelRechargeThread = function(self)
        local status = self.IntelStatus
        if status then
            local ticks = 10 * (self.Blueprint.Intel.ReactivateTime or 1)
            --- display progress
            for k = 1, ticks do
                -- prevent changing work progress when we are doing work (such as an enhancement)
                if not self.WorkItem then
                    self:SetWorkProgress((k / ticks))
                end

                WaitTicks(1)
            end
            -- prevent changing work progress when we are doing work (such as an enhancement)
            if not self.WorkItem then
                self:SetWorkProgress(-1)
            end
            self:OnIntelRecharged()
        end
    end,

    ---@param self IntelComponent | Unit
    ---@param intel? IntelType
    OnIntelEnabled = function(self, intel)
        if myDebug then WARN("Enabled intel: " .. tostring(intel)) end
        if intel == 'Cloak' or intel == 'CloakField' then
            self:UpdateCloakEffect(true, intel)
        end
    end,

    ---@param self IntelComponent | Unit
    ---@param intel? IntelType
    OnIntelDisabled = function(self, intel)
        if myDebug then WARN("Disabled intel: " .. tostring(intel)) end
        if intel == 'Cloak' or intel == 'CloakField' then
            self:UpdateCloakEffect(false, intel)
        end
    end,

    ---@param self IntelComponent | Unit
    ---@param cloaked boolean
    ---@param intel IntelType
    UpdateCloakEffect = function(self, cloaked, intel)
        -- When debugging cloak FX issues, remember that once a structure unit is seen by the enemy,
        -- recloaking won't make it vanish again, and they'll see the new FX.
        if self and not self.Dead then
            if intel == 'Cloak' then
                local bpDisplay = self.Blueprint.Display

                if cloaked then
                    self:SetMesh(bpDisplay.CloakMeshBlueprint, true)
                else
                    self:SetMesh(bpDisplay.MeshBlueprint, true)
                end
            elseif intel == 'CloakField' then
                if self.CloakFieldWatcherThread then
                    KillThread(self.CloakFieldWatcherThread)
                    self.CloakFieldWatcherThread = nil
                end

                if cloaked then
                    self.CloakFieldWatcherThread = self:ForkThread(self.CloakFieldWatcher)
                end
            end
        end
    end,

    ---@param self IntelComponent | Unit
    CloakFieldWatcher = function(self)
        if self and not self.Dead then
            local bp = self.Blueprint
            local radius = bp.Intel.CloakFieldRadius - 2 -- Need to take off 2, because engine reasons
            local brain = self:GetAIBrain()

            while self and not self.Dead and self:IsIntelEnabled('CloakField') do
                local pos = self:GetPosition()
                local units = brain:GetUnitsAroundPoint(categories.ALLUNITS, pos, radius, 'Ally')

                for _, unit in units do
                    if unit and not unit.Dead and unit ~= self then
                        if unit.CloakFXWatcherThread then
                            KillThread(unit.CloakFXWatcherThread)
                            unit.CloakFXWatcherThread = nil
                        end

                        unit:UpdateCloakEffect(true, 'Cloak') -- Turn on the FX for the unit
                        unit.CloakFXWatcherThread = unit:ForkThread(unit.CloakFXWatcher)
                    end
                end
                WaitTicks(6)
            end
        end
    end,

    ---@param self IntelComponent | Unit
    CloakFXWatcher = function(self)
        WaitTicks(6)
        if self and not self.Dead then
            self:UpdateCloakEffect(false, 'Cloak')
        end
    end,


    ---@param self Unit
    ---@param hook fun(unit: Unit, army: number)
    AddDetectedByHook = function(self, hook)
        if not self.DetectedByHooks then
            self.DetectedByHooks = {}
        end
        table.insert(self.DetectedByHooks, hook)
    end,

    ---@param self Unit
    ---@param hook fun(unit: Unit, army: number)
    RemoveDetectedByHook = function(self, hook)
        if self.DetectedByHooks then
            for k, v in self.DetectedByHooks do
                if v == hook then
                    table.remove(self.DetectedByHooks, k)
                    return
                end
            end
        end
    end,

    ---@param self Unit
    ---@param index integer
    OnDetectedBy = function(self, index)
        if self.DetectedByHooks then
            for k, v in self.DetectedByHooks do
                v(self, index)
            end
        end
    end,
}


